from __future__ import annotations

from datetime import date
from typing import Literal
from typing import Optional
from typing import Union

from dac7.models.enums import CountryCode
from dac7.models.enums import CurrencyCode
from dac7.models.enums import EUCountryCode
from dac7.models.enums import LegalAddressType
from dac7.models.enums import NamePersonType
from dac7.models.enums import Nexus
from dac7.models.enums import OtherIdentificationNumberType
from dac7.models.enums import ProdDocSpecType
from dac7.models.enums import TestDocSpecType

import pydantic


class DpiBase(pydantic.BaseModel):
    pass


class DpiNamePersonItem(DpiBase):
    type: Optional[str] = pydantic.Field(default=None, min_length=1, max_length=200, alias="@xnlNameType")
    value: str = pydantic.Field(min_length=1, max_length=200, alias="$")


class DpiNamePerson(DpiBase):
    type: NamePersonType = pydantic.Field(alias="@nameType")
    titles: list[str] = pydantic.Field(default_factory=list, alias="dpi:Title")
    first_name: DpiNamePersonItem = pydantic.Field(alias="dpi:FirstName")
    middle_names: list[DpiNamePersonItem] = pydantic.Field(default_factory=list, alias="dpi:MiddleName")
    name_prefix: Optional[DpiNamePersonItem] = pydantic.Field(default=None, alias="dpi:NamePrefix")
    last_name: DpiNamePersonItem = pydantic.Field(alias="dpi:LastName")


class DpiTaxIdentificationNumber(DpiBase):
    unknown: bool = pydantic.Field(default=False, alias="@unknown")
    issued_by: Optional[CountryCode] = pydantic.Field(default=None, alias="@issuedBy")
    value: str = pydantic.Field(min_length=1, max_length=200, alias="$")


class DpiAddressFix(DpiBase):
    street: str = pydantic.Field(min_length=1, max_length=200, alias="dpi:Street")
    building_identifier: Optional[str] = pydantic.Field(
        default=None,
        min_length=1,
        max_length=200,
        alias="dpi:BuildingIdentifier",
    )
    floor_identifier: Optional[str] = pydantic.Field(
        default=None,
        min_length=1,
        max_length=200,
        alias="dpi:FloorIdentifier",
    )
    post_code: str = pydantic.Field(min_length=1, max_length=200, alias="dpi:PostCode")
    city: str = pydantic.Field(pattern=r".*(\S).*", min_length=1, max_length=200, alias="dpi:City")


class DpiAddress(DpiBase):
    legal_type: Optional[LegalAddressType] = pydantic.Field(default=None, alias="@legalAddressType")
    country_code: CountryCode = pydantic.Field(alias="dpi:CountryCode")
    address_fix: DpiAddressFix = pydantic.Field(alias="dpi:AddressFix")


class DpiBirthPlaceCountryInfo(DpiBase):
    country_code: CountryCode = pydantic.Field(alias="dpi:CountryCode")


class DpiBirthPlace(DpiBase):
    city: str = pydantic.Field(min_length=1, max_length=200, alias="dpi:City")
    city_subentity: Optional[str] = pydantic.Field(
        default=None,
        min_length=1,
        max_length=200,
        alias="dpi:CitySubentity",
    )
    country_info: DpiBirthPlaceCountryInfo = pydantic.Field(alias="dpi:CountryInfo")


class DpiBirthInfo(DpiBase):
    birth_date: date = pydantic.Field(alias="dpi:BirthDate")
    birth_place: Optional[DpiBirthPlace] = pydantic.Field(default=None, alias="dpi:BirthPlace")


class DpiPersonParty(DpiBase):
    residence_country_code: EUCountryCode = pydantic.Field(alias="dpi:ResCountryCode")
    tax_identification_numbers: list[DpiTaxIdentificationNumber] = pydantic.Field(alias="dpi:TIN", min_length=1)
    vat_number: Optional[str] = pydantic.Field(default=None, min_length=1, max_length=200, alias="dpi:VAT")
    name: DpiNamePerson = pydantic.Field(alias="dpi:Name")
    addresses: list[DpiAddress] = pydantic.Field(alias="dpi:Address", min_length=1)
    birth_info: DpiBirthInfo = pydantic.Field(alias="dpi:BirthInfo")

    @pydantic.field_validator("tax_identification_numbers")
    @classmethod
    def check_tax_identification_numbers(
        cls, tax_identification_numbers: list[DpiTaxIdentificationNumber], info: pydantic.ValidationInfo
    ) -> list[DpiTaxIdentificationNumber]:
        unknown_tins = [tin for tin in tax_identification_numbers if tin.unknown]
        if len(tax_identification_numbers) > len(unknown_tins) > 0:
            raise ValueError("Unknown TIN mixed with known TINs")
        return tax_identification_numbers

    @pydantic.model_validator(mode="after")
    def check_tax_identification_numbers_and_birth_place(self) -> DpiPersonParty:
        if self.birth_info.birth_place is not None:
            return self
        if self.tax_identification_numbers[0].unknown:
            raise ValueError("TIN or birth place is mandatory")
        return self


class DpiMonetaryAmount(DpiBase):
    currency_code: CurrencyCode = pydantic.Field(alias="@currCode")
    amount: pydantic.NonNegativeInt = pydantic.Field(alias="$")


class DpiNameOrganization(DpiBase):
    type: Optional[NamePersonType] = pydantic.Field(default=None, alias="@nameType")
    value: str = pydantic.Field(pattern=r".*(\S).*", min_length=1, max_length=200, alias="$")


class DpiMessageSpec(DpiBase):
    transmitting_country: Literal["FR"] = pydantic.Field(alias="dpi:TransmittingCountry")
    receiving_country: Literal["FR"] = pydantic.Field(alias="dpi:ReceivingCountry")
    message_type: Literal["DPI"] = pydantic.Field(alias="dpi:MessageType")
    message_ref_id: str = pydantic.Field(
        pattern=r"^OP_(202[3-9]|20[3-9]\d)_(\d{9})_.+$",
        min_length=1,
        max_length=88,
        alias="dpi:MessageRefId",
    )
    message_type_indic: str = pydantic.Field(alias="dpi:MessageTypeIndic")
    reporting_period: str = pydantic.Field(alias="dpi:ReportingPeriod")
    timestamp: str = pydantic.Field(alias="dpi:Timestamp")


class DpiOtherIdentificationNumber(DpiBase):
    issued_by: CountryCode = pydantic.Field(alias="@issuedBy")
    type: OtherIdentificationNumberType = pydantic.Field(alias="@INType")
    value: str = pydantic.Field(pattern=r".*(\S).*", min_length=1, max_length=200, alias="$")


class DpiOrganizationParty(DpiBase):
    residence_country_code: EUCountryCode = pydantic.Field(alias="dpi:ResCountryCode")
    tax_identification_numbers: list[DpiTaxIdentificationNumber] = pydantic.Field(alias="dpi:TIN")
    other_identification_numbers: list[DpiOtherIdentificationNumber] = pydantic.Field(
        default_factory=list, alias="dpi:IN"
    )
    vat_number: Optional[str] = pydantic.Field(default=None, min_length=1, max_length=200, alias="dpi:VAT")
    legal_names: list[DpiNameOrganization] = pydantic.Field(alias="dpi:Name")
    addresses: list[DpiAddress] = pydantic.Field(alias="dpi:Address", min_length=1)


class DpiDocSpec(DpiBase):
    doc_type_indic: Union[ProdDocSpecType, TestDocSpecType] = pydantic.Field(alias="stf:DocTypeIndic")
    doc_ref_id: str = pydantic.Field(
        pattern=r"[\da-zA-Z+-.:=_]{1,200}",
        min_length=1,
        max_length=200,
        alias="stf:DocRefId",
    )
    corr_doc_ref_id: Optional[str] = pydantic.Field(
        default=None,
        pattern=r"[\da-zA-Z+-.:=_]{1,200}",
        min_length=1,
        max_length=200,
        alias="stf:CorrDocRefId",
    )


class DpiCorrectablePlatformOperator(DpiOrganizationParty):
    business_names: list[str] = pydantic.Field(min_length=1, max_length=200, alias="dpi:PlatformBusinessName")
    nexus: Nexus = pydantic.Field(alias="dpi:Nexus")
    assumed_reporting: bool = pydantic.Field(alias="dpi:AssumedReporting")
    doc_spec: DpiDocSpec = pydantic.Field(alias="dpi:DocSpec")


class DpiIdentifier(DpiBase):
    account_number_type: Optional[str] = pydantic.Field(
        default=None,
        min_length=1,
        max_length=200,
        alias="@AccountNumberType",
    )
    value: str = pydantic.Field(min_length=1, max_length=200, alias="$")


class DpiFinancialIdentifier(DpiBase):
    identifier: DpiIdentifier = pydantic.Field(alias="dpi:Identifier")
    account_holder_name: Optional[str] = pydantic.Field(
        default=None,
        min_length=1,
        max_length=200,
        alias="dpi:AccountHolderName",
    )
    other_info: Optional[str] = pydantic.Field(default=None, min_length=1, max_length=400, alias="dpi:OtherInfo")


class DpiConsideration(DpiBase):
    consideration_q1: DpiMonetaryAmount = pydantic.Field(alias="dpi:ConsQ1")
    consideration_q2: DpiMonetaryAmount = pydantic.Field(alias="dpi:ConsQ2")
    consideration_q3: DpiMonetaryAmount = pydantic.Field(alias="dpi:ConsQ3")
    consideration_q4: DpiMonetaryAmount = pydantic.Field(alias="dpi:ConsQ4")

    @pydantic.model_validator(mode="after")
    def check_consideration_is_positive(self) -> DpiConsideration:
        total = (
            self.consideration_q1.amount
            + self.consideration_q2.amount
            + self.consideration_q3.amount
            + self.consideration_q4.amount
        )
        if total == 0:
            raise ValueError("Total consideration is null")
        return self


class DpiNumberOfActivities(DpiBase):
    activities_number_q1: pydantic.NonNegativeInt = pydantic.Field(alias="dpi:NumbQ1")
    activities_number_q2: pydantic.NonNegativeInt = pydantic.Field(alias="dpi:NumbQ2")
    activities_number_q3: pydantic.NonNegativeInt = pydantic.Field(alias="dpi:NumbQ3")
    activities_number_q4: pydantic.NonNegativeInt = pydantic.Field(alias="dpi:NumbQ4")


class DpiFees(DpiBase):
    fees_q1: DpiMonetaryAmount = pydantic.Field(alias="dpi:FeesQ1")
    fees_q2: DpiMonetaryAmount = pydantic.Field(alias="dpi:FeesQ2")
    fees_q3: DpiMonetaryAmount = pydantic.Field(alias="dpi:FeesQ3")
    fees_q4: DpiMonetaryAmount = pydantic.Field(alias="dpi:FeesQ4")


class DpiTaxes(DpiBase):
    taxes_q1: DpiMonetaryAmount = pydantic.Field(alias="dpi:TaxesQ1")
    taxes_q2: DpiMonetaryAmount = pydantic.Field(alias="dpi:TaxesQ2")
    taxes_q3: DpiMonetaryAmount = pydantic.Field(alias="dpi:TaxesQ3")
    taxes_q4: DpiMonetaryAmount = pydantic.Field(alias="dpi:TaxesQ4")


class DpiOtherActivities(DpiBase):
    consideration: DpiConsideration = pydantic.Field(alias="dpi:Consideration")
    number_of_activities: DpiNumberOfActivities = pydantic.Field(alias="dpi:NumberOfActivities")
    fees: DpiFees = pydantic.Field(alias="dpi:Fees")
    taxes: DpiTaxes = pydantic.Field(alias="dpi:Taxes")


class DpiPropertyListing(DpiOtherActivities):
    address: DpiAddress = pydantic.Field(alias="dpi:Address")
    land_registration_number: str = pydantic.Field(alias="dpi:LandRegistrationNumber")
    property_type: str = pydantic.Field(alias="dpi:PropertyType")
    rented_days: int = pydantic.Field(alias="dpi:RentedDays")


class DpiPermanentEstablishments(DpiBase):
    permanent_establishments: list[EUCountryCode] = pydantic.Field(alias="dpi:PermanentEstablishment")


class DpiStandardEntitySeller(DpiBase):
    entity_seller_details: DpiOrganizationParty = pydantic.Field(alias="dpi:EntSellerID")
    financial_identifiers: list[DpiFinancialIdentifier] = pydantic.Field(
        default_factory=list,
        alias="dpi:FinancialIdentifier",
    )
    permanent_establishments: Optional[DpiPermanentEstablishments] = pydantic.Field(
        default=None,
        alias="dpi:PermanentEstablishments",
    )


class DpiStandardIndividualSeller(DpiBase):
    individual_seller_details: DpiPersonParty = pydantic.Field(alias="dpi:IndSellerID")
    financial_identifiers: list[DpiFinancialIdentifier] = pydantic.Field(
        default_factory=list,
        alias="dpi:FinancialIdentifier",
    )


class DpiEntitySeller(DpiBase):
    standard: DpiStandardEntitySeller = pydantic.Field(alias="dpi:Standard")


class DpiIndividualSeller(DpiBase):
    standard: DpiStandardIndividualSeller = pydantic.Field(alias="dpi:Standard")


class DpiIdentity(DpiBase):
    entity_seller: Optional[DpiEntitySeller] = pydantic.Field(default=None, alias="dpi:EntitySeller")
    individual_seller: Optional[DpiIndividualSeller] = pydantic.Field(default=None, alias="dpi:IndividualSeller")


class DpiImmovableProperty(DpiBase):
    property_listings: list[DpiPropertyListing] = pydantic.Field(alias="dpi:PropertyListing")


class DpiRelevantActivities(DpiBase):
    immovable_property: Optional[DpiImmovableProperty] = pydantic.Field(default=None, alias="dpi:ImmovableProperty")
    personal_services: Optional[DpiOtherActivities] = pydantic.Field(default=None, alias="dpi:PersonalServices")
    sale_of_goods: Optional[DpiOtherActivities] = pydantic.Field(default=None, alias="dpi:SaleOfGoods")
    transportation_rental: Optional[DpiOtherActivities] = pydantic.Field(default=None, alias="dpi:TransportationRental")

    @pydantic.model_validator(mode="after")
    def check_activities(self) -> DpiRelevantActivities:
        for activity in ["immovable_property", "personal_services", "sale_of_goods", "transportation_rental"]:
            if getattr(self, activity) is not None:
                return self
        raise ValueError("No activity")


class DpiReportableSeller(DpiBase):
    identity: DpiIdentity = pydantic.Field(alias="dpi:Identity")
    relevant_activities: DpiRelevantActivities = pydantic.Field(alias="dpi:RelevantActivities")


class DpiCorrectableReportableSeller(DpiReportableSeller):
    doc_spec: DpiDocSpec = pydantic.Field(alias="dpi:DocSpec")


class DpiOtherReportablePlatformOperator(DpiBase):
    residence_country_code: list[CountryCode] = pydantic.Field(alias="dpi:ResCountryCode")
    tax_identification_numbers: list[DpiTaxIdentificationNumber] = pydantic.Field(alias="dpi:TIN")
    legal_name: DpiNameOrganization = pydantic.Field(alias="dpi:Name")
    address: DpiAddress = pydantic.Field(alias="dpi:Address")


class DpiCorrectableOtherReportablePlatformOperator(DpiOtherReportablePlatformOperator):
    doc_spec: DpiDocSpec = pydantic.Field(alias="dpi:DocSpec")


class DpiOtherPlatformOperators(DpiBase):
    assuming_platform_operator: Optional[DpiCorrectableOtherReportablePlatformOperator] = pydantic.Field(
        default=None,
        alias="dpi:AssumingPlatformOperator",
    )
    assumed_platform_operators: list[DpiCorrectableOtherReportablePlatformOperator] = pydantic.Field(
        default_factory=list,
        alias="dpi:AssumedPlatformOperator",
    )


class DpiBody(DpiBase):
    platform_operator: DpiCorrectablePlatformOperator = pydantic.Field(alias="dpi:PlatformOperator")
    other_platform_operators: Optional[DpiOtherPlatformOperators] = pydantic.Field(
        default=None,
        alias="dpi:OtherPlatformOperators",
    )
    reportable_sellers: list[DpiCorrectableReportableSeller] = pydantic.Field(
        default_factory=list,
        alias="dpi:ReportableSeller",
    )


class DpiDeclaration(DpiBase):
    dpi_namespace: Literal["urn:oecd:ties:dpi"] = pydantic.Field(alias="@xmlns:dpi")
    stf_namespace: Literal["urn:oecd:ties:dpistf"] = pydantic.Field(alias="@xmlns:stf")
    schema_version: Literal["1.0"] = pydantic.Field(alias="@version")
    message_spec: DpiMessageSpec = pydantic.Field(alias="dpi:MessageSpec")
    body: DpiBody = pydantic.Field(alias="dpi:DPIBody")



import gzip
import json

from datetime import UTC
from datetime import datetime
from pathlib import Path
from typing import Annotated
from typing import Any
from typing import Optional
from typing import cast

from dac7.constants import Env
from dac7.constants import FileFormat
from dac7.encryption import EncryptionService
from dac7.encryption import KeyInfo
from dac7.models.dpi import DpiDeclaration
from dac7.models.flat import Declaration
from dac7.models.flat import OtherPlatformOperators
from dac7.models.flat import PlatformOperator
from dac7.models.flat import ReportableEntitySeller
from dac7.models.flat import ReportableIndividualSeller
from dac7.naming import build_filename
from dac7.naming import validate_filename

import typer
import xmlschema

# from dateutil.parser import parser

app = typer.Typer()

DAC7_SCHEMA = Path(__file__).parent / "schemas" / "DPIXML_v1.1-fr1.xsd"


@app.command()
def validate(
    xml_path: Annotated[
        Path,
        typer.Argument(
            metavar="XML_FILE",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            show_default=False,
        ),
    ],
    schema_path: Annotated[
        Path,
        typer.Option(
            "--xsd",
            "-x",
            metavar="SCHEMA",
            envvar="DAC7_SCHEMA",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="XSD schema to use.",
            show_default="built-in schema",
        ),
    ] = DAC7_SCHEMA,
) -> None:
    """
    Validate a XML file.

    Check that both name and contents conform to the DAC7 specification.
    """

    schema = xmlschema.XMLSchema10(schema_path)

    with typer.open_file(f"{xml_path}", mode="r") as xml_file:
        xml_data: str = xml_file.read()

    schema.validate(xml_data)

    xml_declaration = cast(dict[str, dict[str, str]], schema.decode(xml_data))
    message_ref_id = xml_declaration["dpi:MessageSpec"]["dpi:MessageRefId"]
    timestamp = xml_declaration["dpi:MessageSpec"]["dpi:Timestamp"]
    validate_filename(xml_path, message_ref_id=message_ref_id, timestamp=timestamp)


@app.command()
def name(
    xml_path: Annotated[
        Path,
        typer.Argument(
            metavar="XML_FILE",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            show_default=False,
        ),
    ],
    file_format: Annotated[
        FileFormat,
        typer.Option(
            "--format",
            "-f",
            help="Format of the file.",
        ),
    ] = FileFormat.XML,
    schema_path: Annotated[
        Path,
        typer.Option(
            "--xsd",
            "-x",
            metavar="SCHEMA",
            envvar="DAC7_SCHEMA",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="XSD schema to use.",
            show_default="built-in schema",
        ),
    ] = DAC7_SCHEMA,
) -> None:
    """
    Return the expected name of the declaration file.
    """

    schema = xmlschema.XMLSchema10(schema_path)

    with typer.open_file(f"{xml_path}", mode="r") as xml_file:
        xml_data: str = xml_file.read()

    schema.validate(xml_data)

    xml_declaration = cast(dict[str, dict[str, str]], schema.decode(xml_data))
    message_ref_id = xml_declaration["dpi:MessageSpec"]["dpi:MessageRefId"]
    timestamp = xml_declaration["dpi:MessageSpec"]["dpi:Timestamp"]

    expected_filename = build_filename(message_ref_id=message_ref_id, timestamp=timestamp)

    typer.echo(f"{expected_filename}.{file_format.value.lower()}")


@app.command()
def build(
    env: Annotated[
        Env,
        typer.Option(
            "--env",
            "-E",
            envvar="ENV",
            help="Environment for which the file is built.",
            show_default=False,
        ),
    ],
    platform_path: Annotated[
        Path,
        typer.Option(
            "--platform-operator",
            "-p",
            metavar="PLATFORM_JSON",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="JSON file with the designated platform operator data",
            show_default=False,
        ),
    ],
    other_platforms_path: Annotated[
        Optional[Path],
        typer.Option(
            "--other-platform-operators",
            "-o",
            metavar="OTHER_PLATFORMS_JSON",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="JSON file with the assuming or assumed platforms data",
            show_default=False,
        ),
    ] = None,
    entity_sellers_path: Annotated[
        Optional[Path],
        typer.Option(
            "--entity-sellers",
            "-e",
            metavar="ENTITIES_JSON",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="JSON file with the entity sellers data",
            show_default=False,
        ),
    ] = None,
    individual_sellers_path: Annotated[
        Optional[Path],
        typer.Option(
            "--individual-sellers",
            "-i",
            metavar="INDIVIDUALS_JSON",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="JSON file with the individual sellers data",
            show_default=False,
        ),
    ] = None,
    declaration_id: Annotated[
        int,
        typer.Option(
            "--declaration-id",
            "-d",
            metavar="ID",
            help="Unique id or serial number.",
        ),
    ] = 1,
    fiscal_year: Annotated[
        int,
        typer.Option(
            "--fiscal-year",
            "-y",
            metavar="YEAR",
            help="Fiscal year",
        ),
    ] = datetime.now(UTC).year
    - 1,
    timestamp: Annotated[
        Optional[datetime],
        typer.Option(
            "--timestamp",
            "-t",
            formats=[r"%Y-%m-%dT%H:%M:%S.%f"],
            help="Timestamp of the declaration.",
            show_default="now",
        ),
    ] = None,
    output_format: Annotated[
        FileFormat,
        typer.Option(
            "--format",
            "-f",
            help="Format of the output.",
        ),
    ] = FileFormat.XML,
    output_file_path: Annotated[
        Optional[Path],
        typer.Argument(
            metavar="[OUTPUT_FILE]",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            show_default="stdout",
        ),
    ] = None,
) -> None:
    """
    Build a XML or JSON declaration from simple, flat JSON files.

    The expected schemas for the JSON input files are available, see: dac7 schemas build --help.
    """

    platform_operator = PlatformOperator.model_validate_json(platform_path.read_text())

    other_platform_data = load_json(other_platforms_path, default={})
    entity_sellers_data = load_json(entity_sellers_path, default=[])
    individual_sellers_data = load_json(individual_sellers_path, default=[])

    declaration = Declaration(
        fiscal_year=fiscal_year,
        declaration_id=declaration_id,
        timestamp=timestamp or get_timestamp(),
        platform_operator=platform_operator,
        other_platform_operators=other_platform_data,
        reportable_entity_sellers=entity_sellers_data,
        reportable_individual_sellers=individual_sellers_data,
        env=env,
    )

    dpi_declaration = declaration.get_dpi()

    dpi_data = dpi_declaration.model_dump(by_alias=True, exclude_defaults=True, mode="json")
    dpi_json = json.dumps(dpi_data, indent=4, ensure_ascii=True)
    result = dpi_json.strip()

    if output_format == FileFormat.XML:
        result = json_to_xml(result, schema_path=DAC7_SCHEMA)

    with typer.open_file(f"{output_file_path or '-'}", mode="w") as output_file:
        output_file.write(result)


@app.command()
def encrypt(
    env: Annotated[
        Env,
        typer.Option(
            "--env",
            "-E",
            envvar="ENV",
            help="Environment for which the file is encrypted.",
            show_default=False,
        ),
    ],
    input_file_path: Annotated[
        Path,
        typer.Argument(
            metavar="INPUT_FILE",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            show_default=False,
        ),
    ],
    compression_requested: Annotated[
        bool,
        typer.Option(
            "--gzip",
            "-z",
            help="Compress the input file with GZIP before encrypting it.",
        ),
    ] = False,
    output_file_path: Annotated[
        Optional[Path],
        typer.Argument(
            metavar="[OUTPUT_FILE]",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            show_default="stdout",
        ),
    ] = None,
) -> None:
    """
    Encrypt a DAC7 file, optionally after GZIP compression.

    Requires GnuPG to be installed.
    """

    # Read

    with typer.open_file(f"{input_file_path}", mode="rb") as input_file:
        input_data: bytes = input_file.read()

    # Compress

    if compression_requested:
        input_data = gzip.compress(input_data)

    # Encrypt

    key_info = KeyInfo.for_env(env)
    service = EncryptionService(key_info=key_info)

    result = service.encrypt_data(input_data)

    # Write

    with typer.open_file(f"{output_file_path or '-'}", mode="wb") as output_file:
        output_file.write(result)


# Development tools


@app.command()
def json2xml(
    json_path: Annotated[
        Path,
        typer.Argument(
            metavar="JSON_FILE",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    schema_path: Annotated[
        Path,
        typer.Option(
            "--xsd",
            "-x",
            metavar="SCHEMA",
            envvar="DAC7_SCHEMA",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="XSD schema to use.",
            show_default="built-in schema",
        ),
    ] = DAC7_SCHEMA,
) -> None:
    """
    Transform a JSON file following the DAC7 schema into XML.
    """

    with typer.open_file(f"{json_path}") as json_file:
        json_data: str = json_file.read()

    DpiDeclaration.model_validate_json(json_data)

    result = json_to_xml(json_data, schema_path=schema_path)
    typer.echo(result)


@app.command()
def xml2json(
    xml_path: Annotated[
        Path,
        typer.Argument(
            metavar="XML_FILE",
            allow_dash=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    schema_path: Annotated[
        Path,
        typer.Option(
            "--xsd",
            "-x",
            metavar="SCHEMA",
            envvar="DAC7_SCHEMA",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            help="XSD schema to use.",
            show_default="built-in schema",
        ),
    ] = DAC7_SCHEMA,
) -> None:
    """
    Read an XML file and output it as JSON.
    """
    xml_schema = xmlschema.XMLSchema(schema_path)
    json_data = xmlschema.to_dict(xml_path, schema=xml_schema)
    typer.echo(json.dumps(json_data, indent=4))


schemas_app = typer.Typer()
app.add_typer(
    schemas_app,
    name="schemas",
    help="Display JSON schemas expected by the tool.",
)


@schemas_app.command("json2xml")
def show_json2xml_schema():
    """
    JSON schema expected by json2xml command.
    """
    typer.echo(json.dumps(DpiDeclaration.model_json_schema(), indent=4))


build_schemas_app = typer.Typer()
schemas_app.add_typer(
    build_schemas_app,
    name="build",
    help="JSON schemas expected by build command.",
)


@build_schemas_app.command("platform-operator")
def show_platform_schema():
    """
    JSON schema for --platform-operator input.
    """
    typer.echo(json.dumps(PlatformOperator.model_json_schema(), indent=4))


@build_schemas_app.command("other-platform-operators")
def show_others_schema():
    """
    JSON schema for --other-platform-operators input.
    """
    typer.echo(json.dumps(OtherPlatformOperators.model_json_schema(), indent=4))


@build_schemas_app.command("entity-sellers")
def show_entity_sellers_schema():
    """
    JSON schema for --entity-sellers input.
    """
    typer.echo(json.dumps(ReportableEntitySeller.model_json_schema(), indent=4))


@build_schemas_app.command("individual-sellers")
def show_individual_sellers_schema():
    """
    JSON schema for --individual-sellers input.
    """
    typer.echo(json.dumps(ReportableIndividualSeller.model_json_schema(), indent=4))


def get_timestamp() -> datetime:
    return datetime.now(UTC).replace(tzinfo=None)


def load_json(path: Optional[Path], default: Any) -> Any:
    if path is None:
        return default
    return json.loads(path.read_text())


def json_to_xml(json_data: str, schema_path: Path) -> str:
    xml_schema = xmlschema.XMLSchema10(schema_path)

    xml_data = xmlschema.from_json(json_data, schema=xml_schema, converter=xmlschema.UnorderedConverter)
    xml_content: str = xmlschema.etree_tostring(  # type: ignore[assignment]
        xml_data,  # type: ignore[arg-type]
        namespaces={"dpi": "urn:oecd:ties:dpi", "stf": "urn:oecd:ties:dpistf"},
        encoding="unicode",
    )
    return f'<?xml version="1.0" encoding="UTF-8"?>\n{xml_content}\n'



from __future__ import annotations

from enum import EnumMeta
from enum import StrEnum

from dac7.constants import Env


class MessageSpecType(StrEnum):
    NEW_DECLARATION = "DPI401"
    CORRECTIVE_DECLARATION = "DPI402"
    EMPTY_DECLARATION = "DPI403"


class OtherIdentificationNumberType(StrEnum):
    BUSINESS_REGISTRATION_NUMBER = "BRN"
    INDIVIDUAL_IDENTIFICATION_NUMBER = "IIN"
    LEGAL_ENTITY_IDENTIFIER = "LEI"
    ENTITY_IDENTIFICATION_NUMBER = "EIN"
    OTHER = "Other"


class NamePersonType(StrEnum):
    # ALIAS_OR_OTHER = "OECD201"
    INDIVIDUAL = "OECD202"
    ALIAS = "OECD203"
    NICKNAME = "OECD204"
    ALSO_KNOWN_AS = "OECD205"
    DOING_BUSINESS_AS = "OECD206"
    LEGAL = "OECD207"
    AT_BIRTH = "OECD208"


class LegalAddressType(StrEnum):
    RESIDENTIAL_OR_BUSINESS = "OECD301"
    RESIDENTIAL = "OECD302"
    BUSINESS = "OECD303"
    REGISTERED_OFFICE = "OECD304"
    UNSPECIFIED = "OECD305"


class Nexus(StrEnum):
    RESIDENT = "RPONEX1"
    INCORPORATED = "RPONEX2"
    MANAGED = "RPONEX3"
    ESTABLISHED = "RPONEX4"
    FACILITATOR = "RPONEX5"


class DocSpecTypeMeta(EnumMeta):
    DATA_RESENT: StrEnum
    NEW_DATA: StrEnum
    DATA_CORRECTED: StrEnum
    DATA_DELETION: StrEnum


class DocSpecType(StrEnum, metaclass=DocSpecTypeMeta):
    @classmethod
    def for_env(cls, env: Env) -> DocSpecTypeMeta:
        return ProdDocSpecType if env == Env.PROD else TestDocSpecType


class ProdDocSpecType(DocSpecType):
    DATA_RESENT = "OECD0"
    NEW_DATA = "OECD1"
    DATA_CORRECTED = "OECD2"
    DATA_DELETION = "OECD3"


class TestDocSpecType(DocSpecType):
    DATA_RESENT = "OECD10"
    NEW_DATA = "OECD11"
    DATA_CORRECTED = "OECD12"
    DATA_DELETION = "OECD13"


class EUCountryCode(StrEnum):
    AUSTRIA = "AT"
    BELGIUM = "BE"
    BULGARIA = "BG"
    CYPRUS = "CY"
    CZECHIA = "CZ"
    DENMARK = "DK"
    ESTONIA = "EE"
    FINLAND = "FI"
    FRANCE = "FR"
    GERMANY = "DE"
    GREECE = "GR"
    HUNGARY = "HU"
    CROATIA = "HR"
    IRELAND = "IE"
    ITALY = "IT"
    LATVIA = "LV"
    LITUANIA = "LT"
    LUXEMBURG = "LU"
    MALTA = "MT"
    NETHERLANDS = "NL"
    POLAND = "PL"
    PORTUGAL = "PT"
    ROMANIA = "RO"
    SLOVAKIA = "SK"
    SLOVENIA = "SI"
    SPAIN = "ES"
    SWEDEN = "SE"


class CountryCode(StrEnum):
    AFGHANISTAN = "AF"
    # ALAND_ISLANDS = "AX"
    ALBANIA = "AL"
    ALGERIA = "DZ"
    AMERICAN_SAMOA = "AS"
    ANDORRA = "AD"
    ANGOLA = "AO"
    ANGUILLA = "AI"
    ANTARCTICA = "AQ"
    ANTIGUA_AND_BARBUDA = "AG"
    ARGENTINA = "AR"
    ARMENIA = "AM"
    ARUBA = "AW"
    AUSTRALIA = "AU"
    AUSTRIA = "AT"
    AZERBAIJAN = "AZ"
    BAHAMAS = "BS"
    BAHRAIN = "BH"
    BANGLADESH = "BD"
    BARBADOS = "BB"
    BELARUS = "BY"
    BELGIUM = "BE"
    BELIZE = "BZ"
    BENIN = "BJ"
    BERMUDA = "BM"
    BHUTAN = "BT"
    BOLIVIA = "BO"
    BONAIRE = "BQ"
    BOSNIA_AND_HERZEGOVINA = "BA"
    BOTSWANA = "BW"
    BOUVET_ISLAND = "BV"
    BRAZIL = "BR"
    BRITISH_INDIAN_OCEAN_TERRITORY = "IO"
    BRUNEI_DARUSSALAM = "BN"
    BULGARIA = "BG"
    BURKINA_FASO = "BF"
    BURUNDI = "BI"
    CAMBODIA = "KH"
    CAMEROON = "CM"
    CANADA = "CA"
    CABO_VERDE = "CV"
    CAYMAN_ISLANDS = "KY"
    CENTRAL_AFRICAN_REPUBLIC = "CF"
    CHAD = "TD"
    CHILE = "CL"
    CHINA = "CN"
    CHRISTMAS_ISLAND = "CX"
    COCOS_ISLANDS = "CC"
    COLOMBIA = "CO"
    COMOROS = "KM"
    CONGO_BRAZZAVILLE = "CG"
    CONGO_KINSHASA = "CD"
    COOK_ISLANDS = "CK"
    COSTA_RICA = "CR"
    IVORY_COAST = "CI"
    CROATIA = "HR"
    CUBA = "CU"
    CURACAO = "CW"
    CYPRUS = "CY"
    CZECHIA = "CZ"
    DENMARK = "DK"
    DJIBOUTI = "DJ"
    DOMINICA = "DM"
    DOMINICAN_REPUBLIC = "DO"
    ECUADOR = "EC"
    EGYPT = "EG"
    EL_SALVADOR = "SV"
    EQUATORIAL_GUINEA = "GQ"
    ERITREA = "ER"
    ESTONIA = "EE"
    ETHIOPIA = "ET"
    FALKLAND_ISLANDS = "FK"
    FAROE_ISLANDS = "FO"
    FIJI = "FJ"
    FINLAND = "FI"
    FRANCE = "FR"
    FRENCH_GUIANA = "GF"
    FRENCH_POLYNESIA = "PF"
    FRENCH_SOUTHERN_TERRITORIES = "TF"
    GABON = "GA"
    GAMBIA = "GM"
    GEORGIA = "GE"
    GERMANY = "DE"
    GHANA = "GH"
    GIBRALTAR = "GI"
    GREECE = "GR"
    GREENLAND = "GL"
    GRENADA = "GD"
    # GUADELOUPE = "GP"
    GUAM = "GU"
    GUATEMALA = "GT"
    GUERNSEY = "GG"
    GUINEA = "GN"
    GUINEA_BISSAU = "GW"
    # GUYANA = "GY"
    HAITI = "HT"
    HEARD_AND_MCDONALD_ISLANDS = "HM"
    VATICAN = "VA"
    HONDURAS = "HN"
    HONG_KONG = "HK"
    HUNGARY = "HU"
    ICELAND = "IS"
    INDIA = "IN"
    INDONESIA = "ID"
    IRAN = "IR"
    IRAQ = "IQ"
    IRELAND = "IE"
    ISLE_OF_MAN = "IM"
    ISRAEL = "IL"
    ITALY = "IT"
    JAMAICA = "JM"
    JAPAN = "JP"
    JERSEY = "JE"
    JORDAN = "JO"
    KAZAKHSTAN = "KZ"
    KENYA = "KE"
    KIRIBATI = "KI"
    NORTH_KOREA = "KP"
    SOUTH_KOREA = "KR"
    KUWAIT = "KW"
    KYRGYZSTAN = "KG"
    LAOS = "LA"
    LATVIA = "LV"
    LEBANON = "LB"
    LESOTHO = "LS"
    LIBERIA = "LR"
    LIBYA = "LY"
    LIECHTENSTEIN = "LI"
    LITHUANIA = "LT"
    LUXEMBOURG = "LU"
    MACAO = "MO"
    NORTH_MACEDONIA = "MK"
    MADAGASCAR = "MG"
    MALAWI = "MW"
    MALAYSIA = "MY"
    MALDIVES = "MV"
    MALI = "ML"
    MALTA = "MT"
    MARSHALL_ISLANDS = "MH"
    # MARTINIQUE = "MQ"
    MAURITANIA = "MR"
    MAURITIUS = "MU"
    # MAYOTTE = "YT"
    MEXICO = "MX"
    MICRONESIA = "FM"
    MOLDOVA = "MD"
    MONACO = "MC"
    MONGOLIA = "MN"
    MONTENEGRO = "ME"
    MONTSERRAT = "MS"
    MOROCCO = "MA"
    MOZAMBIQUE = "MZ"
    MYANMAR = "MM"
    NAMIBIA = "NA"
    NAURU = "NR"
    NEPAL = "NP"
    NETHERLANDS = "NL"
    NEW_CALEDONIA = "NC"
    NEW_ZEALAND = "NZ"
    NICARAGUA = "NI"
    NIGER = "NE"
    NIGERIA = "NG"
    NIUE = "NU"
    NORFOLK_ISLAND = "NF"
    NORTHERN_MARIANA_ISLANDS = "MP"
    NORWAY = "NO"
    OMAN = "OM"
    PAKISTAN = "PK"
    PALAU = "PW"
    PALESTINE = "PS"
    PANAMA = "PA"
    PAPUA_NEW_GUINEA = "PG"
    PARAGUAY = "PY"
    PERU = "PE"
    PHILIPPINES = "PH"
    PITCAIRN = "PN"
    POLAND = "PL"
    PORTUGAL = "PT"
    PUERTO_RICO = "PR"
    QATAR = "QA"
    # REUNION = "RE"
    ROMANIA = "RO"
    RUSSIA = "RU"
    RWANDA = "RW"
    SAINT_BARTHELEMY = "BL"
    SAINT_HELENA_ASCENSION_AND_TRISTAN_DA_CUNHA = "SH"
    SAINT_KITTS_AND_NEVIS = "KN"
    SAINT_LUCIA = "LC"
    SAINT_MARTIN = "MF"
    SAINT_PIERRE_AND_MIQUELON = "PM"
    SAINT_VINCENT_AND_THE_GRENADINES = "VC"
    SAMOA = "WS"
    SAN_MARINO = "SM"
    SAO_TOME_AND_PRINCIPE = "ST"
    SAUDI_ARABIA = "SA"
    SENEGAL = "SN"
    SERBIA = "RS"
    SEYCHELLES = "SC"
    SIERRA_LEONE = "SL"
    SINGAPORE = "SG"
    SINT_MAARTEN = "SX"
    SLOVAKIA = "SK"
    SLOVENIA = "SI"
    SOLOMON_ISLANDS = "SB"
    SOMALIA = "SO"
    SOUTH_AFRICA = "ZA"
    SOUTH_GEORGIA_AND_THE_SOUTH_SANDWICH_ISLANDS = "GS"
    SOUTH_SUDAN = "SS"
    SPAIN = "ES"
    SRI_LANKA = "LK"
    SUDAN = "SD"
    SURINAME = "SR"
    SVALBARD_AND_JAN_MAYEN = "SJ"
    ESWATINI = "SZ"
    SWEDEN = "SE"
    SWITZERLAND = "CH"
    SYRIA = "SY"
    TAIWAN = "TW"
    TAJIKISTAN = "TJ"
    TANZANIA = "TZ"
    THAILAND = "TH"
    TIMOR_LESTE = "TL"
    TOGO = "TG"
    TOKELAU = "TK"
    TONGA = "TO"
    TRINIDAD_AND_TOBAGO = "TT"
    TUNISIA = "TN"
    TURKEY = "TR"
    TURKMENISTAN = "TM"
    TURKS_AND_CAICOS_ISLANDS = "TC"
    TUVALU = "TV"
    UGANDA = "UG"
    UKRAINE = "UA"
    UNITED_ARAB_EMIRATES = "AE"
    UNITED_KINGDOM = "GB"
    UNITED_STATES = "US"
    UNITED_STATES_MINOR_OUTLYING_ISLANDS = "UM"
    URUGUAY = "UY"
    UZBEKISTAN = "UZ"
    VANUATU = "VU"
    VENEZUELA = "VE"
    VIET_NAM = "VN"
    BRITISH_VIRGIN_ISLANDS = "VG"
    US_VIRGIN_ISLANDS = "VI"
    WALLIS_AND_FUTUNA = "WF"
    WESTERN_SAHARA = "EH"
    YEMEN = "YE"
    ZAMBIA = "ZM"
    ZIMBABWE = "ZW"
    KOSOVO = "XK"
    # STATELESS = "X5"
    # OTHER_COUNTRY = "XX"


class CurrencyCode(StrEnum):
    UAE_DIRHAM = "AED"
    AFGHANI = "AFN"
    LEK = "ALL"
    ARMENIAN_DRAM = "AMD"
    NETHERLANDS_ANTILLEAN_GUILDER = "ANG"
    KWANZA = "AOA"
    ARGENTINE_PESO = "ARS"
    AUSTRALIAN_DOLLAR = "AUD"
    ARUBAN_FLORIN = "AWG"
    AZERBAIJAN_MANAT = "AZN"
    CONVERTIBLE_MARK = "BAM"
    BARBADOS_DOLLAR = "BBD"
    TAKA = "BDT"
    BULGARIAN_LEV = "BGN"
    BAHRAINI_DINAR = "BHD"
    BURUNDI_FRANC = "BIF"
    BERMUDIAN_DOLLAR = "BMD"
    BRUNEI_DOLLAR = "BND"
    BOLIVIANO = "BOB"
    MVDOL = "BOV"
    BRAZILIAN_REAL = "BRL"
    BAHAMIAN_DOLLAR = "BSD"
    NGULTRUM = "BTN"
    PULA = "BWP"
    BELARUSIAN_RUBLE = "BYN"
    BELARUSSIAN_RUBLE = "BYR"
    BELIZE_DOLLAR = "BZD"
    CANADIAN_DOLLAR = "CAD"
    CONGOLESE_FRANC = "CDF"
    WIR_EURO = "CHE"
    SWISS_FRANC = "CHF"
    WIR_FRANC = "CHW"
    UNIDAD_DE_FOMENTO = "CLF"
    CHILEAN_PESO = "CLP"
    YUAN_RENMINBI = "CNY"
    COLOMBIAN_PESO = "COP"
    UNIDAD_DE_VALOR_REAL = "COU"
    COSTA_RICAN_COLON = "CRC"
    PESO_CONVERTIBLE = "CUC"
    CUBAN_PESO = "CUP"
    CABO_VERDE_ESCUDO = "CVE"
    CZECH_KORUNA = "CZK"
    DJIBOUTI_FRANC = "DJF"
    DANISH_KRONE = "DKK"
    DOMINICAN_PESO = "DOP"
    ALGERIAN_DINAR = "DZD"
    EGYPTIAN_POUND = "EGP"
    NAKFA = "ERN"
    ETHIOPIAN_BIRR = "ETB"
    EURO = "EUR"
    FIJI_DOLLAR = "FJD"
    FALKLAND_ISLANDS_POUND = "FKP"
    POUND_STERLING = "GBP"
    LARI = "GEL"
    GHANA_CEDI = "GHS"
    GIBRALTAR_POUND = "GIP"
    DALASI = "GMD"
    GUINEAN_FRANC = "GNF"
    QUETZAL = "GTQ"
    GUYANA_DOLLAR = "GYD"
    HONG_KONG_DOLLAR = "HKD"
    LEMPIRA = "HNL"
    KUNA = "HRK"
    GOURDE = "HTG"
    FORINT = "HUF"
    RUPIAH = "IDR"
    NEW_ISRAELI_SHEQEL = "ILS"
    INDIAN_RUPEE = "INR"
    IRAQI_DINAR = "IQD"
    IRANIAN_RIAL = "IRR"
    ICELAND_KRONA = "ISK"
    JAMAICAN_DOLLAR = "JMD"
    JORDANIAN_DINAR = "JOD"
    YEN = "JPY"
    KENYAN_SHILLING = "KES"
    SOM = "KGS"
    RIEL = "KHR"
    COMORIAN_FRANC_ = "KMF"
    NORTH_KOREAN_WON = "KPW"
    WON = "KRW"
    KUWAITI_DINAR = "KWD"
    CAYMAN_ISLANDS_DOLLAR = "KYD"
    TENGE = "KZT"
    LAO_KIP = "LAK"
    LEBANESE_POUND = "LBP"
    SRI_LANKA_RUPEE = "LKR"
    LIBERIAN_DOLLAR = "LRD"
    LOTI = "LSL"
    LITHUANIAN_LITAS = "LTL"
    LATVIAN_LATS = "LVL"
    LIBYAN_DINAR = "LYD"
    MOROCCAN_DIRHAM = "MAD"
    MOLDOVAN_LEU = "MDL"
    MALAGASY_ARIARY = "MGA"
    DENAR = "MKD"
    KYAT = "MMK"
    TUGRIK = "MNT"
    PATACA = "MOP"
    HISTORIC_OUGUIYA = "MRO"
    OUGUIYA = "MRU"
    MAURITIUS_RUPEE = "MUR"
    RUFIYAA = "MVR"
    MALAWI_KWACHA = "MWK"
    MEXICAN_PESO = "MXN"
    MEXICAN_UNIDAD_DE_INVERSION = "MXV"
    MALAYSIAN_RINGGIT = "MYR"
    MOZAMBIQUE_METICAL = "MZN"
    NAMIBIA_DOLLAR = "NAD"
    NAIRA = "NGN"
    CORDOBA_ORO = "NIO"
    NORWEGIAN_KRONE = "NOK"
    NEPALESE_RUPEE = "NPR"
    NEW_ZEALAND_DOLLAR = "NZD"
    RIAL_OMANI = "OMR"
    BALBOA = "PAB"
    SOL = "PEN"
    KINA = "PGK"
    PHILIPPINE_PESO = "PHP"
    PAKISTAN_RUPEE = "PKR"
    ZLOTY = "PLN"
    GUARANI = "PYG"
    QATARI_RIAL = "QAR"
    ROMANIAN_LEU = "RON"
    SERBIAN_DINAR = "RSD"
    RUSSIAN_RUBLE = "RUB"
    RWANDA_FRANC = "RWF"
    SAUDI_RIYAL = "SAR"
    SOLOMON_ISLANDS_DOLLAR = "SBD"
    SEYCHELLES_RUPEE = "SCR"
    SUDANESE_POUND = "SDG"
    SWEDISH_KRONA = "SEK"
    SINGAPORE_DOLLAR = "SGD"
    SAINT_HELENA_POUND = "SHP"
    LEONE = "SLL"
    SOMALI_SHILLING = "SOS"
    SURINAM_DOLLAR = "SRD"
    SOUTH_SUDANESE_POUND = "SSP"
    HISTORIC_DOBRA = "STD"
    DOBRA = "STN"
    EL_SALVADOR_COLON = "SVC"
    SYRIAN_POUND = "SYP"
    LILANGENI = "SZL"
    BAHT = "THB"
    SOMONI = "TJS"
    TURKMENISTAN_NEW_MANAT = "TMT"
    TUNISIAN_DINAR = "TND"
    PAANGA = "TOP"
    TURKISH_LIRA = "TRY"
    TRINIDAD_AND_TOBAGO_DOLLAR = "TTD"
    NEW_TAIWAN_DOLLAR = "TWD"
    TANZANIAN_SHILLING = "TZS"
    HRYVNIA = "UAH"
    UGANDA_SHILLING = "UGX"
    US_DOLLAR = "USD"
    US_DOLLAR_NEXT_DAY = "USN"
    US_DOLLAR_SAME_DAY = "USS"
    URUGUAY_PESO_EN_UNIDADES_INDEXADAS = "UYI"
    PESO_URUGUAYO = "UYU"
    UNIDAD_PREVISIONAL = "UYW"
    UZBEKISTAN_SUM = "UZS"
    BOLIVAR = "VEF"
    BOLIVAR_SOBERANO = "VES"
    DONG = "VND"
    VATU = "VUV"
    TALA = "WST"
    CFA_FRANC = "XAF"
    SILVER = "XAG"
    GOLD = "XAU"
    EAST_CARIBBEAN_DOLLAR = "XCD"
    SPECIAL_DRAWING_RIGHT = "XDR"
    CFA_FRANC_BCEAO = "XOF"
    PALLADIUM = "XPD"
    CFP_FRANC = "XPF"
    PLATINUM = "XPT"
    SUCRE = "XSU"
    NO_CURRENCY = "XXX"
    YEMENI_RIAL = "YER"
    RAND = "ZAR"
    ZAMBIAN_KWACHA = "ZMW"
    ZIMBABWE_DOLLAR = "ZWL"



from __future__ import annotations

from hashlib import sha512
from io import BytesIO
from zipfile import ZipFile

from dac7.constants import Env

import requests

from gnupg import GPG


class KeyInfo:
    fingerprint: str
    archive_url: str
    archive_checksum: str

    @classmethod
    def for_env(cls, env: Env) -> KeyInfo:
        return KeyInfoProd() if env == Env.PROD else KeyInfoTest()


class KeyInfoProd(KeyInfo):
    fingerprint = "3AE282C69675932A19D3245BD155B53F9EFF7A61"
    archive_url = "https://www.impots.gouv.fr/sites/default/files/media/1_metier/3_partenaire/tiers_declarants/cdc_td_bilateral/cle_publique_chiffrement_dgfip_tiersdeclarants_prod.zip"
    archive_checksum = (
        "c3350b1db0a7e9536892254fcd796e702f9dad0be12f97032c32349405bc5b0a1ca868"
        "6997240f686e3e302678791644577924ccb72b236ad1df2431a78b66a8"
    )


class KeyInfoTest(KeyInfo):
    fingerprint = "E7125CD404D160E6C1C773299182BFA4B2FCE419"
    archive_url = "https://www.impots.gouv.fr/sites/default/files/media/1_metier/3_partenaire/tiers_declarants/cdc_td_bilateral/cle_publique_chiffrement_dgfip_tiersdeclarants_test.zip"
    archive_checksum = (
        "ec7f4ec691f6b8a846969b95d8c04b4bf699099adca0d56367248bccd64cd8c57f4315"
        "27e10335cc454424b7952cb80a222e4b27a0f89cfdff014425ffdcfd42"
    )


class EncryptionService:

    def __init__(self, key_info: KeyInfo):
        self.gpg = GPG()
        self.key_info = key_info

    def encrypt_data(self, data: bytes) -> bytes:
        self.import_key_to_gpg()
        result = self.gpg.encrypt(
            data,
            recipients=self.key_info.fingerprint,
            always_trust=True,
            armor=False,
        )
        if not result.ok:
            raise Exception(result.stderr)

        return result.data

    def import_key_to_gpg(self) -> None:
        for key_info in self.gpg.list_keys():
            if self.key_info.fingerprint in key_info["fingerprint"]:
                return

        key = self.fetch_key()
        result = self.gpg.import_keys(key, extra_args=["--yes"])
        if not result.count:
            raise Exception(result.stderr)

    def fetch_key(self) -> str:
        key_archive = self.get_key_archive()
        for file_name in key_archive.namelist():
            if file_name.endswith(".asc"):
                return key_archive.read(file_name).decode("utf8")
        raise FileNotFoundError()

    def get_key_archive(self) -> ZipFile:
        response = requests.get(self.key_info.archive_url, timeout=30)
        response.raise_for_status()

        content = response.content
        if sha512(content).hexdigest() != self.key_info.archive_checksum:
            raise ValueError()

        stream = BytesIO(content)
        return ZipFile(stream)



from __future__ import annotations

import re

from abc import ABC
from datetime import date
from datetime import datetime
from decimal import ROUND_HALF_UP
from decimal import Decimal
from typing import Annotated
from typing import Optional
from typing import Protocol
from typing import Sequence
from typing import Union
from typing import cast
from typing import overload

from dac7.constants import Env
from dac7.models.dpi import DpiAddress
from dac7.models.dpi import DpiBase
from dac7.models.dpi import DpiBirthPlace
from dac7.models.dpi import DpiConsideration
from dac7.models.dpi import DpiCorrectableOtherReportablePlatformOperator
from dac7.models.dpi import DpiCorrectablePlatformOperator
from dac7.models.dpi import DpiCorrectableReportableSeller
from dac7.models.dpi import DpiDeclaration
from dac7.models.dpi import DpiDocSpec
from dac7.models.dpi import DpiFees
from dac7.models.dpi import DpiFinancialIdentifier
from dac7.models.dpi import DpiImmovableProperty
from dac7.models.dpi import DpiNamePerson
from dac7.models.dpi import DpiNamePersonItem
from dac7.models.dpi import DpiNumberOfActivities
from dac7.models.dpi import DpiOtherActivities
from dac7.models.dpi import DpiOtherIdentificationNumber
from dac7.models.dpi import DpiOtherPlatformOperators
from dac7.models.dpi import DpiPermanentEstablishments
from dac7.models.dpi import DpiPropertyListing
from dac7.models.dpi import DpiRelevantActivities
from dac7.models.dpi import DpiTaxes
from dac7.models.dpi import DpiTaxIdentificationNumber
from dac7.models.enums import CountryCode
from dac7.models.enums import CurrencyCode
from dac7.models.enums import DocSpecType
from dac7.models.enums import EUCountryCode
from dac7.models.enums import LegalAddressType
from dac7.models.enums import MessageSpecType
from dac7.models.enums import NamePersonType
from dac7.models.enums import Nexus
from dac7.models.enums import OtherIdentificationNumberType

import annotated_types
import pydantic

NonNegativeDecimal = Annotated[Decimal, annotated_types.Ge(0)]


class Base(pydantic.BaseModel, ABC):
    model_config = pydantic.ConfigDict(populate_by_name=True)


class TaxIdentificationNumber(Base):
    issued_by: CountryCode
    value: str

    def get_dpi(self) -> DpiTaxIdentificationNumber:
        return DpiTaxIdentificationNumber.model_validate(
            {
                "@issuedBy": self.issued_by,
                "$": self.value,
            }
        )

    @classmethod
    def get_no_tin_dpi(cls) -> DpiTaxIdentificationNumber:
        return DpiTaxIdentificationNumber.model_validate(
            {
                "@unknown": True,
                "$": "NOTIN",
            }
        )


class WithTaxIdentificationNumber(Base):
    tax_identification_numbers: list[TaxIdentificationNumber] = pydantic.Field(default_factory=list)

    def get_tins_dpi(self) -> list[DpiBase]:
        if tin_dpi_list := get_dpi_list(self.tax_identification_numbers):
            return tin_dpi_list
        return [TaxIdentificationNumber.get_no_tin_dpi()]


class MandatoryTaxIdentificationNumber(TaxIdentificationNumber):
    value: str


class OtherIdentificationNumber(Base):
    issued_by: CountryCode
    value: str
    type: OtherIdentificationNumberType

    def get_dpi(self) -> DpiOtherIdentificationNumber:
        return DpiOtherIdentificationNumber.model_validate(
            {
                "@issuedBy": self.issued_by,
                "@INType": self.type,
                "$": self.value,
            }
        )


class WithOtherIdentificationNumber(Base):
    other_identification_numbers: list[OtherIdentificationNumber] = pydantic.Field(default_factory=list)

    def get_oins_dpi(self) -> list[DpiBase]:
        return get_dpi_list(self.other_identification_numbers)


class Address(Base):
    legal_type: Optional[LegalAddressType] = None
    country_code: CountryCode
    street: str
    building_identifier: Optional[str] = None
    floor_identifier: Optional[str] = None
    post_code: str
    city: str

    def get_dpi(self) -> DpiAddress:
        return DpiAddress.model_validate(
            {
                "@legalAddressType": self.legal_type,
                "dpi:CountryCode": self.country_code,
                "dpi:AddressFix": {
                    "dpi:Street": self.street,
                    "dpi:BuildingIdentifier": self.building_identifier,
                    "dpi:FloorIdentifier": self.floor_identifier,
                    "dpi:PostCode": self.post_code,
                    "dpi:City": self.city,
                },
            }
        )


class WithDocSpec(Base):
    updated: bool = True
    version: int = 1

    def get_doc_spec_dpi(self, env: Env, ref_id_prefix: str) -> DpiDocSpec:
        indic_enum = DocSpecType.for_env(env)

        corr_doc_ref_id = None
        doc_type_indic = indic_enum.DATA_RESENT
        doc_ref_id = f"{ref_id_prefix}_v{self.version}"

        if self.updated:
            if self.version > 1:
                doc_type_indic = indic_enum.DATA_CORRECTED
                corr_doc_ref_id = f"{ref_id_prefix}_v{self.version - 1}"
            else:
                doc_type_indic = indic_enum.NEW_DATA

        return DpiDocSpec.model_validate(
            {
                "stf:DocTypeIndic": doc_type_indic,
                "stf:DocRefId": doc_ref_id,
                "stf:CorrDocRefId": corr_doc_ref_id,
            }
        )


class PlatformOperator(WithDocSpec, WithOtherIdentificationNumber):
    residence_country_code: EUCountryCode
    tax_identification_numbers: list[MandatoryTaxIdentificationNumber]
    vat_number: str
    legal_names: list[str]
    business_names: list[str]
    addresses: list[Address]
    nexus: Nexus

    @property
    def main_tax_identification_number(self) -> str:
        return self.tax_identification_numbers[0].value

    def get_platform_dpi(self, declaration: Declaration) -> DpiCorrectablePlatformOperator:
        assumed_reporting = declaration.other_platform_operators.assuming_platform_operator is not None

        return DpiCorrectablePlatformOperator.model_validate(
            {
                "dpi:ResCountryCode": self.residence_country_code,
                "dpi:TIN": get_dpi_list(self.tax_identification_numbers),
                "dpi:IN": self.get_oins_dpi(),
                "dpi:VAT": self.vat_number,
                "dpi:Name": [{"$": name} for name in self.legal_names],
                "dpi:PlatformBusinessName": self.business_names,
                "dpi:Address": get_dpi_list(self.addresses),
                "dpi:Nexus": self.nexus,
                "dpi:AssumedReporting": assumed_reporting,
                "dpi:DocSpec": self.get_doc_spec_dpi(env=declaration.env, ref_id_prefix=declaration.ref_id_prefix),
            }
        )


class FinancialIdentifier(Base):
    type: str
    account_number: Optional[str]
    account_holder_name: Optional[str] = None
    other_info: Optional[str] = None

    def get_dpi(self) -> DpiFinancialIdentifier:
        return DpiFinancialIdentifier.model_validate(
            {
                "dpi:Identifier": {
                    "@AccountNumberType": self.type,
                    "$": self.account_number,
                },
                "dpi:AccountHolderName": self.account_holder_name,
                "dpi:OtherInfo": self.other_info,
            }
        )


class WithConsideration(Base):
    currency_code: CurrencyCode
    consideration_amount_q1: Optional[NonNegativeDecimal]
    consideration_amount_q2: Optional[NonNegativeDecimal]
    consideration_amount_q3: Optional[NonNegativeDecimal]
    consideration_amount_q4: Optional[NonNegativeDecimal]

    def get_consideration_dpi(self) -> DpiConsideration:
        return DpiConsideration.model_validate(
            {
                "dpi:ConsQ1": {
                    "@currCode": self.currency_code,
                    "$": round_amount(self.consideration_amount_q1),
                },
                "dpi:ConsQ2": {
                    "@currCode": self.currency_code,
                    "$": round_amount(self.consideration_amount_q2),
                },
                "dpi:ConsQ3": {
                    "@currCode": self.currency_code,
                    "$": round_amount(self.consideration_amount_q3),
                },
                "dpi:ConsQ4": {
                    "@currCode": self.currency_code,
                    "$": round_amount(self.consideration_amount_q4),
                },
            }
        )


class WithNumberOfActivities(Base):
    activities_number_q1: Optional[int]
    activities_number_q2: Optional[int]
    activities_number_q3: Optional[int]
    activities_number_q4: Optional[int]

    def get_activities_dpi(self) -> DpiNumberOfActivities:
        return DpiNumberOfActivities.model_validate(
            {f"dpi:NumbQ{quarter}": getattr(self, f"activities_number_q{quarter}") or 0 for quarter in [1, 2, 3, 4]}
        )


class WithFees(Base):
    currency_code: CurrencyCode
    fees_amount_q1: Optional[NonNegativeDecimal]
    fees_amount_q2: Optional[NonNegativeDecimal]
    fees_amount_q3: Optional[NonNegativeDecimal]
    fees_amount_q4: Optional[NonNegativeDecimal]

    def get_fees_dpi(self) -> DpiFees:
        return DpiFees.model_validate(
            {
                f"dpi:FeesQ{quarter}": {
                    "@currCode": self.currency_code,
                    "$": round_amount(getattr(self, f"fees_amount_q{quarter}")),
                }
                for quarter in [1, 2, 3, 4]
            }
        )


class WithTaxes(Base):
    currency_code: CurrencyCode
    taxes_amount_q1: Optional[NonNegativeDecimal]
    taxes_amount_q2: Optional[NonNegativeDecimal]
    taxes_amount_q3: Optional[NonNegativeDecimal]
    taxes_amount_q4: Optional[NonNegativeDecimal]

    def get_taxes_dpi(self) -> DpiTaxes:
        return DpiTaxes.model_validate(
            {
                f"dpi:TaxesQ{quarter}": {
                    "@currCode": self.currency_code,
                    "$": round_amount(getattr(self, f"taxes_amount_q{quarter}")),
                }
                for quarter in [1, 2, 3, 4]
            }
        )


def _validate_quarter_consideration(other_activites: OtherActivities, quarter: int) -> Decimal:
    activities_number = getattr(other_activites, f"activities_number_q{quarter}") or 0
    consideration = getattr(other_activites, f"consideration_amount_q{quarter}") or 0
    if activities_number > 0 and consideration == 0:
        raise ValueError(
            f"Consideration for Q{quarter} is null while if the number of activities is positive",
        )
    return cast(NonNegativeDecimal, consideration)


class OtherActivities(WithConsideration, WithNumberOfActivities, WithFees, WithTaxes):
    @pydantic.model_validator(mode="after")
    def ensure_consideration_is_positive_when_needed(self) -> OtherActivities:
        consideration_amount_total = Decimal("0.")

        for quarter in [1, 2, 3, 4]:
            consideration_amount_total += _validate_quarter_consideration(self, quarter=quarter)

        if consideration_amount_total == 0:
            raise ValueError("Total consideration is null or negative")

        return self

    def get_dpi(self) -> DpiOtherActivities:
        return DpiOtherActivities.model_validate(
            {
                "dpi:Consideration": self.get_consideration_dpi(),
                "dpi:NumberOfActivities": self.get_activities_dpi(),
                "dpi:Fees": self.get_fees_dpi(),
                "dpi:Taxes": self.get_taxes_dpi(),
            }
        )


class WithPermanentEstablishments(Base):
    residence_country_code: Optional[EUCountryCode] = None
    establishment_country_codes: list[EUCountryCode] = pydantic.Field(default_factory=list)

    @pydantic.model_validator(mode="after")
    def remove_residence_country_code(self) -> WithPermanentEstablishments:
        self.establishment_country_codes = [
            code for code in self.establishment_country_codes if code != self.residence_country_code
        ]
        return self

    def get_establishment_country_codes_dpi(self) -> Optional[DpiPermanentEstablishments]:
        if not self.establishment_country_codes:
            return None
        return DpiPermanentEstablishments.model_validate(
            {"dpi:PermanentEstablishment": self.establishment_country_codes}
        )


class PropertyListing(OtherActivities):
    address: Address
    land_registration_number: str
    property_type: str
    rented_days: int

    def get_dpi(self) -> DpiPropertyListing:
        return DpiPropertyListing.model_validate(
            {
                "dpi:Address": self.address.get_dpi(),
                "dpi:LandRegistrationNumber": self.land_registration_number,
                "dpi:PropertyType": self.property_type,
                "dpi:RentedDays": self.rented_days,
                "dpi:Consideration": self.get_consideration_dpi(),
                "dpi:NumberOfActivities": self.get_activities_dpi(),
                "dpi:Fees": self.get_fees_dpi(),
                "dpi:Taxes": self.get_taxes_dpi(),
            }
        )


class WithImmovableProperties(Base):
    immovable_properties: list[PropertyListing] = pydantic.Field(default_factory=list)

    def get_immovable_properties_dpi(self) -> Optional[DpiImmovableProperty]:
        if not self.immovable_properties:
            return None
        return DpiImmovableProperty.model_validate(
            {
                "dpi:PropertyListing": get_dpi_list(self.immovable_properties),
            }
        )


class WithReportableActivities(WithImmovableProperties):
    immovable_properties: list[PropertyListing] = pydantic.Field(default_factory=list)
    personal_services: Optional[OtherActivities] = None
    sale_of_goods: Optional[OtherActivities] = None
    transportation_rental: Optional[OtherActivities] = None

    @pydantic.model_validator(mode="after")
    def check_relevant_activities(self) -> WithReportableActivities:
        if (
            not self.immovable_properties
            and not self.personal_services
            and not self.sale_of_goods
            and not self.transportation_rental
        ):
            raise ValueError("Need at least one relevant activity")
        return self

    def get_reportable_activities_dpi(self) -> DpiRelevantActivities:
        personal_services_dpi = get_dpi(self.personal_services)
        sale_of_goods_dpi = get_dpi(self.sale_of_goods)
        transportation_rental_dpi = get_dpi(self.transportation_rental)

        return DpiRelevantActivities.model_validate(
            {
                "dpi:ImmovableProperty": self.get_immovable_properties_dpi(),
                "dpi:PersonalServices": personal_services_dpi,
                "dpi:SaleOfGoods": sale_of_goods_dpi,
                "dpi:TransportationRental": transportation_rental_dpi,
            }
        )


class ReportableEntitySeller(
    WithTaxIdentificationNumber,
    WithOtherIdentificationNumber,
    WithReportableActivities,
    WithDocSpec,
    WithPermanentEstablishments,
):
    seller_id: str
    vat_number: Optional[str] = None
    legal_names: list[str]
    addresses: list[Address] = pydantic.Field(min_length=1)
    financial_identifiers: list[FinancialIdentifier] = pydantic.Field(default_factory=list)

    def get_dpi(self, declaration: Declaration) -> DpiCorrectableReportableSeller:
        reportable_activities_dpi = self.get_reportable_activities_dpi()

        ref_id_prefix = f"{declaration.ref_id_prefix}_{self.seller_id}"

        return DpiCorrectableReportableSeller.model_validate(
            {
                "dpi:Identity": {
                    "dpi:EntitySeller": {
                        "dpi:Standard": {
                            "dpi:EntSellerID": {
                                "dpi:ResCountryCode": self.residence_country_code,
                                "dpi:TIN": self.get_tins_dpi(),
                                "dpi:IN": self.get_oins_dpi(),
                                "dpi:VAT": self.vat_number or None,
                                "dpi:Name": [{"$": name} for name in self.legal_names],
                                "dpi:Address": get_dpi_list(self.addresses),
                            },
                            "dpi:FinancialIdentifier": get_dpi_list(self.financial_identifiers),
                            "dpi:PermanentEstablishments": self.get_establishment_country_codes_dpi(),
                        }
                    },
                },
                "dpi:RelevantActivities": reportable_activities_dpi,
                "dpi:DocSpec": self.get_doc_spec_dpi(env=declaration.env, ref_id_prefix=ref_id_prefix),
            }
        )


class WithName(Base):
    name_type: NamePersonType
    titles: list[str] = pydantic.Field(default_factory=list)
    first_name: str
    first_name_type: Optional[str] = None
    middle_names: list[DpiNamePersonItem] = pydantic.Field(default_factory=list)
    name_prefix: Optional[DpiNamePersonItem] = None
    last_name: str
    last_name_type: Optional[str] = None

    def get_name_dpi(self) -> DpiNamePerson:
        return DpiNamePerson.model_validate(
            {
                "@nameType": self.name_type,
                "dpi:Title": self.titles,
                "dpi:FirstName": {
                    "@xnlNameType": self.first_name_type,
                    "$": self.first_name,
                },
                "dpi:MiddleName": self.middle_names,
                "dpi:NamePrefix": self.name_prefix,
                "dpi:LastName": {
                    "@xnlNameType": self.last_name_type,
                    "$": self.last_name,
                },
            }
        )


class WithBirthPlace(Base):
    birth_city: str
    birth_country_code: CountryCode
    birth_city_subentity: Optional[str] = None

    def get_birth_place_dpi(self) -> DpiBirthPlace:
        return DpiBirthPlace.model_validate(
            {
                "dpi:City": self.birth_city,
                "dpi:CitySubentity": self.birth_city_subentity,
                "dpi:CountryInfo": {
                    "dpi:CountryCode": self.birth_country_code,
                },
            }
        )


class ReportableIndividualSeller(
    WithName,
    WithBirthPlace,
    WithTaxIdentificationNumber,
    WithReportableActivities,
    WithDocSpec,
):
    seller_id: str
    residence_country_code: EUCountryCode
    vat_number: Optional[str] = None
    birth_date: date
    addresses: list[Address]
    financial_identifiers: list[FinancialIdentifier] = pydantic.Field(default_factory=list)

    @pydantic.model_validator(mode="after")
    def check_tax_identification_number_or_birth_place(self) -> ReportableIndividualSeller:
        if self.birth_country_code is not None or self.tax_identification_numbers:
            return self
        raise ValueError("TIN or birth place is mandatory")

    def get_dpi(self, declaration: Declaration) -> DpiCorrectableReportableSeller:
        reportable_activities_dpi = self.get_reportable_activities_dpi()
        birth_place_dpi = self.get_birth_place_dpi()
        ref_id_prefix = f"{declaration.ref_id_prefix}_{self.seller_id}"

        return DpiCorrectableReportableSeller.model_validate(
            {
                "dpi:Identity": {
                    "dpi:IndividualSeller": {
                        "dpi:Standard": {
                            "dpi:IndSellerID": {
                                "dpi:ResCountryCode": self.residence_country_code,
                                "dpi:TIN": self.get_tins_dpi(),
                                "dpi:VAT": self.vat_number,
                                "dpi:Name": self.get_name_dpi(),
                                "dpi:Address": get_dpi_list(self.addresses),
                                "dpi:BirthInfo": {
                                    "dpi:BirthDate": self.birth_date.isoformat(),
                                    "dpi:BirthPlace": birth_place_dpi,
                                },
                            },
                            "dpi:FinancialIdentifier": get_dpi_list(self.financial_identifiers),
                        },
                    }
                },
                "dpi:RelevantActivities": reportable_activities_dpi,
                "dpi:DocSpec": self.get_doc_spec_dpi(env=declaration.env, ref_id_prefix=ref_id_prefix),
            }
        )


class OtherPlatformOperator(WithDocSpec):
    residence_country_codes: list[CountryCode]
    tax_identification_numbers: list[MandatoryTaxIdentificationNumber]
    legal_name: str
    address: Address

    def get_dpi(self, declaration: Declaration) -> DpiCorrectableOtherReportablePlatformOperator:
        main_tin = re.sub(r"[^0-9a-zA-Z+-.:=_]", "", self.tax_identification_numbers[0].value)
        ref_id_prefix = f"{declaration.ref_id_prefix}_{main_tin}"
        return DpiCorrectableOtherReportablePlatformOperator.model_validate(
            {
                "dpi:ResCountryCode": self.residence_country_codes,
                "dpi:TIN": get_dpi_list(self.tax_identification_numbers),
                "dpi:Name": {"$": self.legal_name},
                "dpi:Address": self.address.get_dpi(),
                "dpi:DocSpec": self.get_doc_spec_dpi(env=declaration.env, ref_id_prefix=ref_id_prefix),
            }
        )


class OtherPlatformOperators(Base):
    assuming_platform_operator: Optional[OtherPlatformOperator] = None
    assumed_platform_operators: list[OtherPlatformOperator] = pydantic.Field(default_factory=list)

    def get_dpi(self, declaration: Declaration) -> Optional[DpiOtherPlatformOperators]:
        if not self.assuming_platform_operator and not self.assumed_platform_operators:
            return None

        return DpiOtherPlatformOperators.model_validate(
            {
                "dpi:AssumingPlatformOperator": get_dpi(self.assuming_platform_operator, declaration=declaration),
                "dpi:AssumedPlatformOperator": get_dpi_list(self.assumed_platform_operators, declaration=declaration),
            }
        )


class WithReportableSellers(Base):
    reportable_entity_sellers: list[ReportableEntitySeller] = pydantic.Field(default_factory=list)
    reportable_individual_sellers: list[ReportableIndividualSeller] = pydantic.Field(default_factory=list)

    def get_reportable_sellers_dpi(self, declaration: Declaration) -> list[DpiBase]:
        dpi_entity_sellers = get_dpi_list(self.reportable_entity_sellers, declaration=declaration)
        dpi_individual_sellers = get_dpi_list(self.reportable_individual_sellers, declaration=declaration)
        return dpi_entity_sellers + dpi_individual_sellers


class Declaration(WithReportableSellers):
    fiscal_year: int
    declaration_id: int
    timestamp: datetime
    platform_operator: PlatformOperator
    other_platform_operators: OtherPlatformOperators
    env: Env

    @property
    def ref_id_prefix(self) -> str:
        parts = [
            "OP",
            f"{self.fiscal_year}",
            f"{self.platform_operator.main_tax_identification_number}",
        ]

        if self.env == Env.TEST:
            parts.append(f"t{self.declaration_id}")

        return "_".join(parts)

    def get_dpi(self) -> DpiDeclaration:
        message_type_indic = MessageSpecType.NEW_DECLARATION
        if self.other_platform_operators.assuming_platform_operator:
            message_type_indic = MessageSpecType.EMPTY_DECLARATION

        return DpiDeclaration.model_validate(
            {
                "@xmlns:dpi": "urn:oecd:ties:dpi",
                "@xmlns:stf": "urn:oecd:ties:dpistf",
                "@version": "1.0",
                "dpi:MessageSpec": {
                    "dpi:TransmittingCountry": "FR",
                    "dpi:ReceivingCountry": "FR",
                    "dpi:MessageType": "DPI",
                    "dpi:MessageRefId": f"{self.ref_id_prefix}_d{self.declaration_id}",
                    "dpi:MessageTypeIndic": f"{message_type_indic}",
                    "dpi:ReportingPeriod": f"{self.fiscal_year}-12-31",
                    "dpi:Timestamp": self.timestamp.isoformat()[:-3],
                },
                "dpi:DPIBody": {
                    "dpi:PlatformOperator": self.platform_operator.get_platform_dpi(declaration=self),
                    "dpi:OtherPlatformOperators": self.other_platform_operators.get_dpi(declaration=self),
                    "dpi:ReportableSeller": self.get_reportable_sellers_dpi(declaration=self),
                },
            }
        )


class WithGetDpi(Protocol):
    def get_dpi(self) -> DpiBase: ...


class WithGetDpiUsingDeclaration(Protocol):
    def get_dpi(self, declaration: Declaration) -> DpiBase: ...


@overload
def get_dpi(model: Optional[WithGetDpi]) -> Optional[DpiBase]: ...


@overload
def get_dpi(model: Optional[WithGetDpiUsingDeclaration], declaration: Declaration) -> Optional[DpiBase]: ...


def get_dpi(
    model: Union[Optional[WithGetDpi], Optional[WithGetDpiUsingDeclaration]], declaration: Optional[Declaration] = None
) -> Optional[DpiBase]:
    if model is None:
        return None
    kwargs = {"declaration": declaration} if declaration else {}
    return model.get_dpi(**kwargs)


@overload
def get_dpi_list(models: Sequence[WithGetDpi]) -> list[DpiBase]: ...


@overload
def get_dpi_list(models: Sequence[WithGetDpiUsingDeclaration], declaration: Declaration) -> list[DpiBase]: ...


def get_dpi_list(
    models: Union[Sequence[WithGetDpi], Sequence[WithGetDpiUsingDeclaration]], declaration: Optional[Declaration] = None
) -> list[DpiBase]:
    kwargs = {"declaration": declaration} if declaration else {}
    return [input_value.get_dpi(**kwargs) for input_value in models]


def round_amount(value: Optional[Decimal]) -> Decimal:
    if value is None:
        return Decimal("0.")
    return value.quantize(Decimal("1."), rounding=ROUND_HALF_UP)



from __future__ import annotations

import re

from dataclasses import dataclass
from pathlib import Path
from typing import ClassVar


class FilenameValidationError(Exception):
    pass


def validate_filename(file_path: Path, message_ref_id: str, timestamp: str) -> None:
    filename_vars = FilenameVars.from_filename(file_path.name)
    messagespec_vars = MessageSpecVars.from_messagespec(
        message_ref_id=message_ref_id,
        timestamp=timestamp,
    )

    if filename_vars.declaration_year != messagespec_vars.declaration_year:
        raise FilenameValidationError("Year inconsistent with content of file")

    if filename_vars.platform_id != messagespec_vars.platform_id:
        raise FilenameValidationError("Platform id inconsistent with content of file")

    if filename_vars.timestamp != messagespec_vars.timestamp:
        raise FilenameValidationError("Timestamp inconsistent with content of file")


def build_filename(message_ref_id: str, timestamp: str) -> str:
    ms_vars = MessageSpecVars.from_messagespec(
        message_ref_id=message_ref_id,
        timestamp=timestamp,
    )
    return f"DPIDAC7_{ms_vars.declaration_year}_{ms_vars.platform_id}_{ms_vars.declaration_id}_{ms_vars.timestamp}"


@dataclass
class ExpectedVars:
    declaration_year: str
    platform_id: str
    declaration_id: str
    timestamp: str


class FilenameVars(ExpectedVars):
    REGEX: ClassVar[str] = (
        r"^DPIDAC7_(?P<declaration_year>[0-9]{4})_(?P<platform_id>[0-9]{9})_"
        r"(?P<declaration_id>[0-9]{3})_(?P<timestamp>[0-9]{14}).xml$"
    )

    @classmethod
    def from_filename(cls, filename: str) -> FilenameVars:
        match = re.match(cls.REGEX, filename)
        if not match:
            raise FilenameValidationError("File does not match expected pattern")

        return cls(
            declaration_year=match.group("declaration_year"),
            platform_id=match.group("platform_id"),
            declaration_id=match.group("declaration_id"),
            timestamp=match.group("timestamp"),
        )


class MessageSpecVars(ExpectedVars):
    REGEX: ClassVar[str] = r"^OP_(?P<declaration_year>[0-9]{4})_(?P<platform_id>[0-9]{9})_"

    @classmethod
    def from_messagespec(cls, message_ref_id: str, timestamp: str) -> MessageSpecVars:
        match = re.match(cls.REGEX, message_ref_id)
        if not match:
            raise Exception

        return cls(
            declaration_year=match.group("declaration_year"),
            platform_id=match.group("platform_id"),
            declaration_id="001",
            timestamp=re.sub("[^0-9]", "", timestamp)[:14],
        )
